----------------------------------------------------------------------------------
--  File: depp_cntrl.vhd
----------------------------------------------------------------------------------
--  Author: Jordan Retz
--  CopyRight (c) 2013 Digilent Inc.
----------------------------------------------------------------------------------
--  Revision History:
--  	07/24/2013	Created (Jordan Retz)
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- For IOBUF Instantiations
library UNISIM;
use UNISIM.VCOMPONENTS.ALL;

----------------------------------------------------------------------------------
--	Title: depp_cntrl Entity
----------------------------------------------------------------------------------
--  Inputs: 
--  Outputs: 
----------------------------------------------------------------------------------
--  Description:
--  	
----------------------------------------------------------------------------------

entity depp_cntrl is
    Port ( 
			  -- FPGA_GLK, expecting 8 MHz
			  CLK 				: in  STD_LOGIC;
			  
			  -- DEPP address strobe (ASTB) and DEPP data strobe (DSTB)
           DEPP_ASTB 		: in  STD_LOGIC;
           DEPP_DSTB 		: in  STD_LOGIC;
			  
			  -- DEPP write enable (write operation = 0, read operation = 1)
           DEPP_WRITE 		: in  STD_LOGIC;
			  
			  -- DEPP logic ready 
			  DEPP_WAIT			: out STD_LOGIC;
			  
			  -- DEPP bi-directional data bus. Direction controlled by DEPP_WRITE
           DBUS 				: inout  STD_LOGIC_VECTOR (7 downto 0);
			  
			  -- Internal bus for interfacing with port_io
           INTERNAL_DOUT 	: out  STD_LOGIC_VECTOR (7 downto 0);
			  
			  -- Internal bus for interfacing with port_io
           INTERNAL_DIN 	: in  STD_LOGIC_VECTOR (7 downto 0);
			  
			  -- Internal bus for interfacing with port_io
           INTERNAL_AOUT 	: out  STD_LOGIC_VECTOR (3 downto 0);
			  
			  -- Internal handshake to port_io logic (active high, initial low)
			  INTERNAL_HS		: out STD_LOGIC
			);
end depp_cntrl;

-------------------------------------------------------
-------------------------------------------------------
-- Architecture Start
-------------------------------------------------------
-------------------------------------------------------

architecture Behavioral of depp_cntrl is

-------------------------------------------------------
-- State Constants and variables
-------------------------------------------------------
constant deppSt_Wait : STD_LOGIC_VECTOR (2 downto 0) 	:= "000";		-- Waiting for bus cycle
constant deppSt_Busy : STD_LOGIC_VECTOR (2 downto 0) 	:= "101";		-- Blocking bus cycle
constant deppSt_AW 	: STD_LOGIC_VECTOR (2 downto 0) 	:= "010";		-- Address write bus cycle
constant deppSt_DW 	: STD_LOGIC_VECTOR (2 downto 0) 	:= "001";		-- Data write bus cycle
constant deppSt_DR 	: STD_LOGIC_VECTOR (2 downto 0) 	:= "100";		-- Data read bus cycle

signal deppSt_Cur 	: STD_LOGIC_VECTOR (2 downto 0) 	:= "000";
signal deppSt_Next 	: STD_LOGIC_VECTOR (2 downto 0) 	:= "000";

-------------------------------------------------------
-- DBUS internal signals
-------------------------------------------------------
signal iobuf_dbus_out 	: STD_LOGIC_VECTOR (7 downto 0);
signal iobuf_dbus_in 	: STD_LOGIC_VECTOR (7 downto 0);

-------------------------------------------------------
-- Address Register and handshake signal
-------------------------------------------------------
signal dout_sig		: STD_LOGIC_VECTOR (7 downto 0);
signal addr_pointer 	: STD_LOGIC_VECTOR (3 downto 0) 	:= "0000";
signal hand_shake		: STD_LOGIC								:= '0';

-------------------------------------------------------
-- Address Register and handshake signal
-------------------------------------------------------
signal not_write		: STD_LOGIC := '1'; -- initial value is so that DBUS is input

-------------------------------------------------------
-- Behavioral Begin
-------------------------------------------------------
begin

-------------------------------------------------------
-- IOBUF Instantiations
-------------------------------------------------------
-- Generates 8 IOBUF instances for each DBUS pin
IOBUF_GEN1: for i in 0 to 7 generate
begin
	IOBUFA : IOBUF								
	generic map (
		DRIVE => 12,
		IOSTANDARD => "LVCMOS33",
		SLEW => "SLOW")
	port map (
		O => iobuf_dbus_out(i),
		I => iobuf_dbus_in(i),
		IO => DBUS(i),
		T => not_write	-- ie, DEPP_WRITE == '1' then host read, => IOBUF is set up as output (I active)
	);
end generate;

-------------------------------------------------------
-- State Control
-------------------------------------------------------
process (CLK) begin
	if(CLK'event and CLK = '1') then
		deppSt_Cur <= deppSt_Next;
	end if;
end process;

-------------------------------------------------------
-- Continous Assignments
-------------------------------------------------------
INTERNAL_HS <= hand_shake;
INTERNAL_DOUT <= iobuf_dbus_out;
iobuf_dbus_in <= INTERNAL_DIN;
INTERNAL_AOUT <= addr_pointer;
not_write <= not DEPP_WRITE;

-------------------------------------------------------
-- DEPP State Machine
-------------------------------------------------------

process (DEPP_ASTB, DEPP_DSTB, DEPP_WRITE, deppSt_Cur, deppSt_Next, iobuf_dbus_out) begin

	-- STATE MACHINE
	case deppSt_Cur is
		-- WAIT
		when deppSt_Wait => 
			DEPP_WAIT <= '0'; -- WAIT active low indicates logic is ready for bus cycle
			
			-- Check for state changes
			if (DEPP_ASTB = '0') then									-- address write
				hand_shake <= '0';
				
				deppSt_Next <= deppSt_AW;
				
			elsif (DEPP_DSTB = '0' and DEPP_WRITE = '1') then 	-- data read
				hand_shake <= '1';
				
				deppSt_Next <= deppSt_DR;
			
			elsif (DEPP_DSTB = '0' and DEPP_WRITE = '0') then 	-- data write
				hand_shake <= '1';
				
				deppSt_Next <= deppSt_DW;
				
			else																-- remain in wait state
				hand_shake <= '0';
				
				deppSt_Next <= deppSt_Wait;
			end if;
		
		-- BUSY
		when deppSt_Busy =>
			DEPP_WAIT <= '1';
			hand_shake <= '0';
			
			deppSt_Next <= deppSt_Wait;
		
		-- ADDRESS WRITE
		when deppSt_AW =>
			DEPP_WAIT <= '0';
			addr_pointer <= iobuf_dbus_out(3 downto 0);	-- IOBUF instance is outputting the input data
			hand_shake <= '0';
			
			deppSt_Next <= deppSt_Busy;
					
		-- DATA WRITE
		when deppSt_DW =>
			hand_shake <= '0';
			DEPP_WAIT <= '0';
			
			deppSt_Next <= deppSt_Busy;
		
		-- DATA READ
		when deppSt_DR =>
			hand_shake <= '0';
			DEPP_WAIT <= '0';
			
			deppSt_Next <= deppSt_Busy;
		
		when others =>
			hand_shake <= '0';
			DEPP_WAIT <= '0';
			
			deppSt_Next <= deppSt_Wait;
		
	end case;	

end process;

end Behavioral;

